from fanuc_ucl._fanuc_core import rmi as __rmi

ResponsePacket = __rmi.ResponsePacket  # base response packet class
SendPacket = __rmi.SendPacket  # base send packet class
Command = __rmi.Command  # subclass of SendPacket
CommandResponse = __rmi.CommandResponse  # subclass of ResponsePacket
Instruction = __rmi.Instruction  # subclass of SendPacket
InstructionResponse = __rmi.InstructionResponse  # subclass of ResponsePacket
Communication = __rmi.Communication  # subclass of SendPacket
CommunicationResponse = __rmi.CommunicationResponse  # subclass of ResponsePacket
# subclasses of Command or CommandResponse
FrcInitialize = __rmi.FrcInitialize
FrcAbort = __rmi.FrcAbort
FrcPause = __rmi.FrcPause
FrcContinue = __rmi.FrcContinue
FrcReadError = __rmi.FrcReadError
FrcSetUFrameUTool = __rmi.FrcSetUFrameUTool
FrcGetUFrameUTool = __rmi.FrcGetUFrameUTool
FrcGetStatus = __rmi.FrcGetStatus
FrcReadUFrameData = __rmi.FrcReadUFrameData
FrcWriteUFrameData = __rmi.FrcWriteUFrameData
FrcReadUToolData = __rmi.FrcReadUToolData
FrcWriteUToolData = __rmi.FrcWriteUToolData
FrcReadDIN = __rmi.FrcReadDIN
FrcReadCartesianPosition = __rmi.FrcReadCartesianPosition
FrcWriteDOUT = __rmi.FrcWriteDOUT
FrcReadJointAngles = __rmi.FrcReadJointAngles
FrcSetOverRide = __rmi.FrcSetOverRide
FrcReadPositionRegister = __rmi.FrcReadPositionRegister
FrcWritePositionRegister = __rmi.FrcWritePositionRegister
FrcReset = __rmi.FrcReset
FrcReadTCPSpeed = __rmi.FrcReadTCPSpeed
FrcInitializeResponse = __rmi.FrcInitializeResponse
FrcAbortResponse = __rmi.FrcAbortResponse
FrcPauseResponse = __rmi.FrcPauseResponse
FrcContinueResponse = __rmi.FrcContinueResponse
FrcReadErrorResponse = __rmi.FrcReadErrorResponse
FrcSetUFrameUToolResponse = __rmi.FrcSetUFrameUToolResponse
FrcGetUFrameUToolResponse = __rmi.FrcGetUFrameUToolResponse
FrcGetStatusResponse = __rmi.FrcGetStatusResponse
FrcReadUFrameDataResponse = __rmi.FrcReadUFrameDataResponse
FrcWriteUFrameDataResponse = __rmi.FrcWriteUFrameDataResponse
FrcReadUToolDataResponse = __rmi.FrcReadUToolDataResponse
FrcWriteUToolDataResponse = __rmi.FrcWriteUToolDataResponse
FrcReadDINResponse = __rmi.FrcReadDINResponse
FrcReadCartesianPositionResponse = __rmi.FrcReadCartesianPositionResponse
FrcWriteDOUTResponse = __rmi.FrcWriteDOUTResponse
FrcReadJointAnglesResponse = __rmi.FrcReadJointAnglesResponse
FrcSetOverRideResponse = __rmi.FrcSetOverRideResponse
FrcReadPositionRegisterResponse = __rmi.FrcReadPositionRegisterResponse
FrcWritePositionRegisterResponse = __rmi.FrcWritePositionRegisterResponse
FrcResetResponse = __rmi.FrcResetResponse
FrcReadTCPSpeedResponse = __rmi.FrcReadTCPSpeedResponse
# subclasses of Instruction or InstructionResponse
FrcWaitDIN = __rmi.FrcWaitDIN
FrcSetUFrame = __rmi.FrcSetUFrame
FrcSetUTool = __rmi.FrcSetUTool
FrcWaitTime = __rmi.FrcWaitTime
FrcSetPayLoad = __rmi.FrcSetPayLoad
FrcCall = __rmi.FrcCall
FrcLinearMotion = __rmi.FrcLinearMotion
FrcLinearRelative = __rmi.FrcLinearRelative
FrcLinearRelativeJRep = __rmi.FrcLinearRelativeJRep
FrcJointMotion = __rmi.FrcJointMotion
FrcJointRelative = __rmi.FrcJointRelative
FrcCircularMotion = __rmi.FrcCircularMotion
FrcCircularRelative = __rmi.FrcCircularRelative
FrcJointMotionJRep = __rmi.FrcJointMotionJRep
FrcJointRelativeJRep = __rmi.FrcJointRelativeJRep
FrcLinearMotionJRep = __rmi.FrcLinearMotionJRep
FrcWaitDINResponse = __rmi.FrcWaitDINResponse
FrcSetUFrameResponse = __rmi.FrcSetUFrameResponse
FrcSetUToolResponse = __rmi.FrcSetUToolResponse
FrcWaitTimeResponse = __rmi.FrcWaitTimeResponse
FrcSetPayLoadResponse = __rmi.FrcSetPayLoadResponse
FrcCallResponse = __rmi.FrcCallResponse
FrcLinearMotionResponse = __rmi.FrcLinearMotionResponse
FrcLinearRelativeResponse = __rmi.FrcLinearRelativeResponse
FrcLinearRelativeJRepResponse = __rmi.FrcLinearRelativeJRepResponse
FrcJointMotionResponse = __rmi.FrcJointMotionResponse
FrcJointRelativeResponse = __rmi.FrcJointRelativeResponse
FrcCircularMotionResponse = __rmi.FrcCircularMotionResponse
FrcCircularRelativeResponse = __rmi.FrcCircularRelativeResponse
FrcJointMotionJRepResponse = __rmi.FrcJointMotionJRepResponse
FrcJointRelativeJRepResponse = __rmi.FrcJointRelativeJRepResponse
FrcLinearMotionJRepResponse = __rmi.FrcLinearMotionJRepResponse
# subclasses of Communication or CommunicationResponse
FrcConnect = __rmi.FrcConnect
FrcDisconnect = __rmi.FrcDisconnect
FrcConnectResponse = __rmi.FrcConnectResponse
FrcDisconnectResponse = __rmi.FrcDisconnectResponse
FrcTerminateResponse = __rmi.FrcTerminateResponse
FrcSystemFaultResponse = __rmi.FrcSystemFaultResponse
# member structs
OnOff = __rmi.OnOff
FrameData = __rmi.FrameData
Configuration = __rmi.Configuration
Position = __rmi.Position
JointAngles = __rmi.JointAngles
LcbType = __rmi.LcbType
PortType = __rmi.PortType
LocalConditionBlock = __rmi.LocalConditionBlock
OffsetRegisterNumbers = __rmi.OffsetRegisterNumbers
TermType = __rmi.TermType
SpeedType = __rmi.SpeedType
